#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
#include <stdio.h>
#include <time.h>

static GdkWindow *socket = NULL;
static GtkWidget *window;
static GtkClipboard *pclipboard;
static GtkWidget **buttonArr;
static guchar **buttonStr;
static int buttonArrN=3;
static int maxButtonStrlen=9;
// static GdkAtom atom_cutbuffer0;
static GtkWidget *snoop_button;
static gboolean b_borderless=FALSE,b_autoraise=FALSE;
static char *geomstr=NULL;
GtkTooltips *button_bar_tips;
static GtkWidget *hist_window;
static guchar **hist_strArr;
int hist_strArrN=10;
GtkWidget **hist_buttonArr;



const guchar *make_path(const guchar *dir, const guchar *leaf)
{
static GString *buffer = NULL;
if (!buffer)
     buffer = g_string_new(NULL);
     g_return_val_if_fail(dir != NULL, buffer->str);
     g_return_val_if_fail(leaf != NULL, buffer->str);
     if (buffer->str != dir) 
	g_string_assign(buffer, dir);
     if (dir[0] != '/' || dir[1] != '\0')
	g_string_append_c(buffer, '/'); /* For anything except "/" */
	g_string_append(buffer, leaf);
     return buffer->str;
}







static void del_nl(guchar *tmpstr)
{
   int i;

   for(i=0;tmpstr[i];i++) {
     if (tmpstr[i]=='\n')
       tmpstr[i]=' ';
   }
}

static update_hist_button()
{
  int i;

  for(i=0;i<hist_strArrN;i++) {
    guchar tstr[16];

    if (!hist_strArr[i])
      continue;

    g_snprintf(tstr,sizeof(tstr),"%s",hist_strArr[i]);

    del_nl(tstr);
    gtk_button_set_label(GTK_BUTTON(hist_buttonArr[i]),tstr);
#if 0
    gtk_tooltips_set_tip (GTK_TOOLTIPS (button_bar_tips),hist_buttonArr[i],
      hist_strArr[i],NULL);
#endif
  }
}

static void show_hist_window(gdouble x, gdouble y)
{
  gint x_pos, y_pos;
  GtkAllocation alloc; 

#if 1
  if (geomstr)
    gtk_window_parse_geometry(GTK_WINDOW(hist_window),geomstr);
#endif

    alloc=GTK_WIDGET(hist_window)->allocation;
    x_pos = x;
    y_pos = y;
    if (y_pos>alloc.height) y_pos-=alloc.height;
    gtk_widget_set_uposition (GTK_WIDGET(hist_window),x_pos,y_pos); 

  update_hist_button();

  

  gtk_widget_show (hist_window);
  gtk_window_present(GTK_WINDOW(hist_window));
}


/* Signal handler called when the selections owner returns the data */
static void selection_received(GtkClipboard *pclip, const gchar *text, gpointer data)
{
   guchar *tmpstr;
   GtkWidget *button = (GtkWidget *)data;
   int i;
   int textlen;
   guchar titlestr[34];

   if (!text || !text[0])
     return;

   for(i=0;i<buttonArrN;i++) {
     if (buttonStr[i] && !strcmp(buttonStr[i],text))
       return;
   }

   tmpstr=g_malloc(maxButtonStrlen+1);
#if 0
   // strncpy doesn't work as expected on Mandrake 9.0
   strncpy(tmpstr,text,maxButtonStrlen);
#else
   textlen=strlen(text);
   if (textlen > maxButtonStrlen)
     textlen = maxButtonStrlen;
   memcpy(tmpstr,text,textlen);
   tmpstr[textlen]=0;
#endif

   del_nl(tmpstr);

   for(i=0;i<buttonArrN;i++) {
     if (buttonArr[i]==button) {
       if (buttonStr[i])
         g_free(buttonStr[i]);
       buttonStr[i]=g_strdup(text);
     }
   }

   gtk_button_set_label(GTK_BUTTON(button),tmpstr);

   g_snprintf(titlestr,sizeof(titlestr),"%s",text);

   gtk_window_set_title (GTK_WINDOW (window),titlestr);

   gtk_tooltips_set_tip (GTK_TOOLTIPS (button_bar_tips), button, text,NULL);

   g_free(tmpstr);


//   gtk_window_set_default_size(GTK_WINDOW(button),-1,-1);

   // remove the duplicate item if any
   for(i=0;i< hist_strArrN; i++) {
     if (!hist_strArr[i])
       continue;
     if (strcmp(hist_strArr[i],text))
       continue;

     g_free(hist_strArr[i]);

     memmove(&hist_strArr[i],&hist_strArr[i+1],
             sizeof(hist_strArr[0])*(hist_strArrN- i - 1));

     hist_strArr[hist_strArrN-1]=NULL;
     break;
   }

   g_free(hist_strArr[hist_strArrN-1]);
   memmove(&hist_strArr[1],&hist_strArr[0],
           sizeof(hist_strArr[0])*(hist_strArrN-1));

   hist_strArr[0]=g_strdup(text);

   update_hist_button();
}


void get_selection()
{
  gtk_clipboard_request_text(pclipboard,selection_received,snoop_button);
}

void set_snoop_button(GtkWidget *widget)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget),TRUE);
  snoop_button = widget;
}


static void get_mouse_button( GtkWidget *widget,GdkEventButton *event, gpointer data)
{
  int i;
#if 0
  printf("b gtk_widget_get_events  %d  %x  %d\n",event->type,event->state,event->button);
#endif
  if (event->button == 3) {
    for(i=0;i<buttonArrN;i++) {
      if (buttonArr[i]!=widget)
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(buttonArr[i]),FALSE);
    }

    set_snoop_button(widget);
//    get_selection();
  } else
  if (event->button == 2) {
    show_hist_window(event->x_root,event->y_root);
  } else
  if (event->button == 1) {
    for(i=0;i<buttonArrN;i++) {
      if (buttonArr[i]!=widget)
        continue;
      if (buttonStr[i]) {
        gtk_clipboard_set_text(pclipboard, buttonStr[i], -1);
      }
      break;
    }

  }
}

static void hist_get_mouse_button( GtkWidget *widget,GdkEventButton *event, gpointer data)
{
  int i;
#if 0
  printf("b gtk_widget_get_events  %d  %x  %d\n",event->type,event->state,event->button);
#endif
  if (event->button == 1) {
    for(i=0;i<hist_strArrN;i++) {
      if (hist_buttonArr[i]!=widget)
        continue;
      if (hist_strArr[i]) {
        gtk_clipboard_set_text(pclipboard, hist_strArr[i], -1);

      }
      break;
    }
  }

  gtk_widget_hide(hist_window);
}

gboolean delete_hist_win()
{
  gtk_widget_hide(hist_window);
  return TRUE;
}

static void free_mem()
{
  int i;

  for(i=0;i<buttonArrN;i++)
    g_free(buttonStr[i]);

  g_free(buttonArr);
  g_free(buttonStr);

  for(i=0;i<hist_strArrN;i++)
    g_free(hist_strArr[i]);
  g_free(hist_strArr);
  g_free(hist_buttonArr);
}

static void do_exit()
{
  gtk_main_quit();
}


GdkFilterReturn property_change_event(GdkXEvent *xevent, GdkEvent *event, gpointer user_data)
{
  XPropertyEvent *xprop = (XPropertyEvent *)xevent;
  GdkEventProperty *prop = &event->property;

  if (event->property.window!=GDK_ROOT_PARENT())
    return GDK_FILTER_REMOVE;
#if 0
  if (prop->atom != atom_cutbuffer0 && prop->atom != GDK_SELECTION_PRIMARY &&
      prop->atom != GDK_SELECTION_CLIPBOARD)
    return GDK_FILTER_REMOVE;
#endif
//  printf("property change %d\n",prop->atom);
  fflush(stdout);

  static gint64 last_time;
  gint64 ti;
  struct timeval tval;

  gettimeofday(&tval,NULL);

  ti = (gint64)tval.tv_sec * 1000000 + tval.tv_usec;

  // mozilla sometimes causes a dead lock condition
#if 1
  if (ti - last_time > 100000)
#endif
  {
    last_time = ti;
    get_selection();
  }

  return GDK_FILTER_REMOVE;
}

gboolean  key_press_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
   if ((event->string && event->string[0]=='q') || event->keyval==GDK_Escape)
     do_exit();
#if 0
   printf("event->keyval:%x\n",event->keyval);
#endif
   return TRUE;
}


gboolean  hist_key_press_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
   gtk_widget_hide(hist_window);
//   if ((event->string && event->string[0]=='q') || event->keyval==GDK_Escape)
//   printf("hist event->keyval:%x\n",event->keyval);
   return TRUE;
}


gboolean  win_focus_in(GtkWidget *widget, GdkEventFocus *event,
                       gpointer user_data)
{
   gtk_window_present(GTK_WINDOW(widget));
}


gboolean  button_scroll_event(GtkWidget *widget,GdkEventScroll *event, gpointer user_data)
{
  int winx,winy,i;

  if (event->direction!=GDK_SCROLL_DOWN)
    return TRUE;

  show_hist_window(event->x_root,event->y_root);
}

gboolean hist_focus_out_callback(GtkWidget *widget, GdkEventFocus *event,
                                 gpointer user_data)
{
   gtk_widget_hide(hist_window);
   return TRUE;
}


int main(int argc, char *argv[])
{
  GtkWidget *hbox,*vbox;
  GtkWidget *proxy_invisible;
  int i;
  gint32 xid=0;


  gtk_init (&argc, &argv);

  
  
 for(i=1;i<argc;i++) {
    if (!strcmp(argv[i],"-g") || !strcmp(argv[i],"-geometry")
        || !strcmp(argv[i],"--geometry") && i < argc-1) {
       geomstr=argv[i+1];
       i++;
    } else
    if (!strcmp(argv[i],"-B")) {
      b_borderless=TRUE;
    } else
    if (!strcmp(argv[i],"-ar")) {
      b_autoraise=TRUE;
    } else
    if (!strcmp(argv[i],"-nb")) {
      buttonArrN=atoi(argv[i+1]);
      i++;
    } else
    if (!strcmp(argv[i],"-nhb")) {
      hist_strArrN=atoi(argv[i+1]);
      i++;
    } else
    if (!strcmp(argv[i],"-mbsl")) {
      maxButtonStrlen=atoi(argv[i+1]);
      if (maxButtonStrlen < 2)
        maxButtonStrlen = 2;
      i++;
      } else
    if (!strcmp(argv[i],"-xid")) {
      xid=atoi(argv[i+1]);
      break;
      
    } else {
      fprintf(stderr,
"---------- gtk copy-paste buffer version " VERSION " ---------\n\n"
"\t -g string\t window geometry: (default: %s)\n"
"\t -B\t\t window borderless (no decoration)\n"
"\t -ar\t\t auto raise (raise window when focused in)\n"
"\t -nb number\t number-of-buttons: (default: %d)\n"
"\t -nhb number\t number-of-history-buttons: (default: %d)\n"
"\t -mbsl number\t maximal-button-string-length: (default: %d)\n\n"
"\t Left mouse button to put content to paste-buffer\n"
"\t Right mouse button to select the button as a snooper\n"
"\t Middle mouse button press/scroll to show the history window\n"
"\t Escape or 'q' to quit\n\n",geomstr, buttonArrN,hist_strArrN,maxButtonStrlen);
      exit(-1);
    }
  } 


  if (xid != 0) 
  {
  window = gtk_plug_new(xid);
  socket = gdk_window_foreign_new(xid); 
  }
  else
  {
   window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  }
  

  buttonArr=g_malloc(buttonArrN * sizeof(GtkWidget *));
  buttonStr=g_malloc0(buttonArrN * sizeof(guchar *));

  hist_strArr=g_malloc0(hist_strArrN * sizeof(guchar *));
  hist_buttonArr=g_malloc(hist_strArrN * sizeof(GtkWidget *));


  hist_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

  GtkWidget *win_icon=gtk_image_new_from_file(make_path(g_strdup(getenv("APP_DIR")), ".DirIcon"));

  gtk_window_set_icon(GTK_WINDOW(window),
     gtk_image_get_pixbuf(GTK_IMAGE(win_icon)));
  gtk_window_set_icon(GTK_WINDOW(hist_window),
     gtk_image_get_pixbuf(GTK_IMAGE(win_icon)));

  // Under gnome 2.0, the window is not fixed if decorated, annoying
  gtk_window_set_decorated(GTK_WINDOW(hist_window),FALSE);
  gtk_window_set_title (GTK_WINDOW (hist_window),"gcb history");

  gtk_window_set_title (GTK_WINDOW(window), "gcb: gtk copy-paste buffer");
  gtk_window_stick(GTK_WINDOW(window));

  g_signal_connect(G_OBJECT (window),"destroy", G_CALLBACK (do_exit), NULL);
  g_signal_connect(G_OBJECT (hist_window),"delete_event",
    G_CALLBACK (delete_hist_win), NULL);
  g_signal_connect(G_OBJECT (hist_window),"focus-out-event",
    G_CALLBACK (hist_focus_out_callback), NULL);

  if (b_autoraise) {
    g_signal_connect(G_OBJECT (window),"focus-in-event",
                     G_CALLBACK (win_focus_in), NULL);
  }

  button_bar_tips = gtk_tooltips_new ();

  hbox = gtk_hbox_new (FALSE, 1);
  vbox = gtk_vbox_new (FALSE, 1);

  if (geomstr)
    gtk_window_parse_geometry(GTK_WINDOW(window),geomstr);

  if (xid !=0)
  {
    GtkSeparator *gliph = GTK_SEPARATOR(gtk_vseparator_new());

    gtk_box_pack_start (GTK_BOX(hbox), GTK_WIDGET(gliph), TRUE, TRUE, FALSE);
    gtk_widget_show_all (GTK_WIDGET(hbox));
  }

  for(i=0;i<buttonArrN;i++) {
    buttonArr[i] = gtk_toggle_button_new_with_label ("---");
    gtk_button_set_relief (GTK_BUTTON(buttonArr[i]),GTK_RELIEF_NONE);
    gtk_box_pack_start (GTK_BOX(hbox), buttonArr[i], TRUE, TRUE, FALSE);
    gtk_widget_show (buttonArr[i]);
    g_signal_connect (G_OBJECT (buttonArr[i]), "button-press-event",
                      G_CALLBACK (get_mouse_button), (gpointer) buttonArr[i]);
#if 1
    g_signal_connect (G_OBJECT (buttonArr[i]), "key-press-event",
                      G_CALLBACK (key_press_event), NULL);
    g_signal_connect (G_OBJECT (buttonArr[i]), "scroll-event",
                      G_CALLBACK (button_scroll_event), NULL);
#endif
  }


  for(i=0;i<hist_strArrN;i++) {
    hist_buttonArr[i] = gtk_button_new_with_label ("---");
    gtk_box_pack_start (GTK_BOX(vbox), hist_buttonArr[i], TRUE, TRUE, FALSE);
    gtk_widget_show (hist_buttonArr[i]);
    g_signal_connect (G_OBJECT (hist_buttonArr[i]), "button-press-event",
                      G_CALLBACK (hist_get_mouse_button), (gpointer) hist_buttonArr[i]);
#if 1
    g_signal_connect (G_OBJECT (hist_buttonArr[i]), "key-press-event",
                      G_CALLBACK (hist_key_press_event), NULL);
#endif
  }

  gtk_container_add (GTK_CONTAINER(window), hbox);
  gtk_container_add (GTK_CONTAINER(hist_window), vbox);

  gtk_widget_show (hbox);
  gtk_widget_show (vbox);
  gtk_widget_show (window);
  
  gdk_input_set_extension_events(window->window,
      GDK_EXTENSION_EVENTS_ALL,
      GDK_EXTENSION_EVENTS_ALL);

  pclipboard = gtk_clipboard_get(GDK_SELECTION_PRIMARY);

  set_snoop_button(buttonArr[0]);
  get_selection();

  gdk_window_set_events (GDK_ROOT_PARENT (), GDK_PROPERTY_CHANGE_MASK);
  gdk_window_add_filter (GDK_ROOT_PARENT (), property_change_event, NULL);

//  atom_cutbuffer0 = gdk_x11_xatom_to_atom(XA_CUT_BUFFER0);

  gtk_container_set_border_width(GTK_CONTAINER(hbox),0);
  gtk_container_set_border_width(GTK_CONTAINER(window),0);

  // -B must be first
  if (b_borderless)
    gtk_window_set_decorated(GTK_WINDOW(window),FALSE);

  if (geomstr)
    gtk_window_parse_geometry(GTK_WINDOW(window),geomstr);

  gtk_main ();

  free_mem();

  return 0;
}
